/**************************************************************************/
/*                                                                        */
/* file:         wb2txt.c                                                 */
/* version:      1.1 (April 1, 2005)                                      */
/*                                                                        */
/* description:  This tool detokenizes files that are saved with the      */
/*               assembler WB-ASS2 for MSX.                               */
/*                                                                        */
/*  This program is free software; you can redistribute it and/or modify  */
/*  it under the terms of the GNU General Public License as published by  */
/*  the Free Software Foundation; either version 2, or (at your option)   */
/*  any later version. See COPYING for more details.                      */
/*                                                                        */
/*                                                                        */
/* Copyright 2003 Vincent van Dam (vincentd@kip.sateh.com)                */
/* Copyright 2005 The New Image (http://www.tni.nl/)                      */
/*                                                                        */
/* v1.0 (October 19, 2003) by Vincent van Dam                             */
/* - Initial release                                                      */
/*                                                                        */
/* v1.1 (April 1, 2005) by Patriek Lesparre/The New Image                 */
/* - No more program crash on RLCA                                        */
/* - No TAB before comments on start of line                              */
/* - Changed &H and &B notations to $ and % for compatibility reasons     */
/*                                                                        */
/**************************************************************************/

#include <stdio.h>
#include <stdlib.h>

#ifndef SEEK_SET
#define SEEK_SET 0
#define SEEK_END 2
#endif

#ifndef bool
#define true   1
#define false  0
#define bool   int
#endif

#ifndef uint8
#define uint8 unsigned char
#endif

#ifndef uint16
#define uint16 unsigned int
#endif

/***********************/
/* wbass2 token tables */
/***********************/

char *tokens_1[81] = { "LD","JR","DJNZ","CALL","RET","JP","INC","DEC","PUSH",
		       "POP","RST","IN","OUT","IM","EX","ADD","ADC","SUB",
		       "SBC","AND","XOR","OR","CP","RLC","RRC","RL","RR","SLA",
		       "SRA","???","SRL","BIT","RES","SET","CPD","CPDR","CPI",
		       "CPIR","IND","INDR","INI","INIR","LDD","LDDR","LDI",
		       "LDIR","OUTD","OTDR","OUTI","OTIR","NEG","RETI","RETN",
		       "RLD","RRD","CCF","CPL","DAA","DI","EI","EXX","HALT",
		       "NOP","RLA","RLCA","RRA","RRCA","SCF","ORG","EQU","END",
		       "DB","DW","DS","DM","DEFB","DEFW","DEFS","DEFM",
		       "GLOBAL","INCLUDE" };

char *tokens_2[29] = { "A","B","C","D","E","H","L","I","R","BC","DE","HL","SP",
		       "IX","IY","AF","NZ","Z","NC","C","PO","PE","P","M","$",
		       "AND","XOR","OR","MOD" };

char tokens_3[9] = ",)(+-*/^";

/****************/
/* exit program */
/****************/

void exit_error(msg)
char *msg;
{
  printf("wb2txt: %s\n",msg);
  printf("usage: wb2txt <file>\n");
  exit(1);
}

/****************/
/* let's do it! */
/****************/

int main(argc,argv)
int argc;
char* argv[];
{
  FILE   *wbfile;
  uint8  *buffer;
  uint8  c;
  uint16 size;
  uint16 labels;
  uint16 label;
  uint16 number;
  uint16 i,j,k;
  uint16 endline;
  uint8  length;
  uint8  bits;
  bool   instr;
  char   binbuf[17];

  if (argc>2) exit_error("syntax error");

  /* open file */
  if (argc==1)
    wbfile=stdin; 
  else
    if ((wbfile=fopen(argv[1],"rb"))==NULL) exit_error("cannot open file");

  /* check if it is a wbass2 file */
  if (getc(wbfile)!=0xfd) exit_error("not a wbass2 file");

  /* determine filesize */
  fseek(wbfile,0,SEEK_END);
  size=ftell(wbfile)-1;
  fseek(wbfile,1,SEEK_SET);

  /* read entire file */
  buffer=(uint8*)malloc(size);
  fread(buffer,size,1,wbfile);

  fclose(wbfile);

  /* first run; get location of labels */
  for(i=0;i<size;) {
    c=(buffer[i]+1)&127;
    i=i+c;
    if (!c) break;
  }
  labels = i+1;

  /* second run; detokenize data */
  for(i=0;i<size;i=endline) {
    if(buffer[i]==0xff) break;
    length=buffer[i]; i++;
    endline=i+(length&127);
    /* does the line start with a label? */
    if(length&128) {
      label = buffer[i]; i++;
      label+= buffer[i]*256; i++;
      for(j=labels+8*label;(buffer[j]&127)&&j<(labels+8*label+6);j++)
	printf("%c",buffer[j]&127);
      printf(":");
      length-=128;
    }
    if (buffer[i]!=1) printf("\t");
    /* decode rest of line */
    instr=true;
    for(j=i;j<endline;) {
      /* decimal numbers */
      c=buffer[j++];
      if(c==0xe0) {
	number = buffer[j++];
	number+= buffer[j++]*256;
       	printf("%d",number);
      }
      /* hexadecimal numbers */
      if(c==0xe1) {
	number = buffer[j++];
	number+= buffer[j++]*256;
	if(number>256)
	  printf("$%04X",number);
	else
	  printf("$%02X",number);
      }
      /* binary numbers */
      if(c==0xe2) {
	number = buffer[j++];
	number+= buffer[j++]*256;
	if(number>256)
	  bits=16;
	else
	  bits=8;
	printf("\%");
	for(k=0;k<bits;k++) {
	  if(number&1)
	    binbuf[bits-k-1] ='1';
	  else
	    binbuf[bits-k-1] ='0';
	  number=number>>1;
	}
	binbuf[bits]=0;
	printf("%s",binbuf);
      }
      /* comments */
      if(c==1) {
	if(!instr) printf("\t");
	printf("; ");
	for(;j<endline;j++) printf("%c",buffer[j]);
      }
      /* instructions/operands */
      if(c>127 && c!=0xe0 && c!=0xe1 && c!=0xe2) {
	if(instr) {
	  printf("%s\t",tokens_1[c-128]);
	  instr=false;
	} else {
          /* labels */
          if(c==0xc0) {
	    label = buffer[j++];
	    label+= buffer[j++]*256;
	    for(k=labels+8*label;(buffer[k]&127)&&k<(labels+8*label+6);k++)
	      printf("%c",buffer[k]&127);
          } else {
	    printf("%s",tokens_2[c-128]);
          }
	}
      }
      /* special characters */
      if(c!=1 && c<18)
       	printf("%c",tokens_3[c/2-1]);
      /* text */
      if(c==34) {
	printf("\"");
	do { 
	  c=buffer[j++];
	  printf("%c",c);
	} while (c!=34 && j<endline);
      }
    }
    printf("\n");
  }

  return 0;
}
